/*
* Copyright (c) 2022, NVIDIA CORPORATION. All rights reserved.
*
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions
* are met:
*  * Redistributions of source code must retain the above copyright
*    notice, this list of conditions and the following disclaimer.
*  * Redistributions in binary form must reproduce the above copyright
*    notice, this list of conditions and the following disclaimer in the
*    documentation and/or other materials provided with the distribution.
*  * Neither the name of NVIDIA CORPORATION nor the names of its
*    contributors may be used to endorse or promote products derived
*    from this software without specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS ``AS IS'' AND ANY
* EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
* IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
* PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER OR
* CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
* EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
* PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
* PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
* OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
* (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
* OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

#include <opencv2/core/version.hpp>
#if CV_MAJOR_VERSION >= 3
#    include <opencv2/imgcodecs.hpp>
#else
#    include <opencv2/highgui/highgui.hpp>
#endif
#include <opencv2/imgproc/imgproc.hpp>
#include <vpi/OpenCVInterop.hpp>

#include <vpi/Array.h>
#include <vpi/Image.h>
#include <vpi/Status.h>
#include <vpi/Stream.h>
#include <vpi/algo/ConvertImageFormat.h>
#include <vpi/algo/MinMaxLoc.h>
#include <vpi/algo/TemplateMatching.h>

#include <cassert>
#include <cstring> // for memset
#include <iostream>
#include <sstream>

#define CHECK_STATUS(STMT)                                    \
    do                                                        \
    {                                                         \
        VPIStatus status = (STMT);                            \
        if (status != VPI_SUCCESS)                            \
        {                                                     \
            char buffer[VPI_MAX_STATUS_MESSAGE_LENGTH];       \
            vpiGetLastStatusMessage(buffer, sizeof(buffer));  \
            std::ostringstream ss;                            \
            ss << vpiStatusGetName(status) << ": " << buffer; \
            throw std::runtime_error(ss.str());               \
        }                                                     \
    } while (0);

int main(int argc, char *argv[])
{
    // OpenCV image that will be wrapped by a VPIImage.
    // Define it here so that it's destroyed *after* wrapper is destroyed
    cv::Mat cvImage;
    cv::Mat cvImageU8;

    cv::Mat cvTempl;

    VPIStream stream = NULL;

    // VPI objects that will be used
    VPIImage input = NULL;
    VPIImage templ = NULL;
    int originX, originY, templWidth, templHeight;

    int outWidth, outHeight;
    VPIImage output       = NULL;
    VPIImage outputScaled = NULL;
    VPIImage outputU8     = NULL;

    VPIArray minCoords = NULL;
    VPIArray maxCoords = NULL;

    VPIPayload payload       = NULL;
    VPIPayload payloadMinMax = NULL;

    int retval = 0;

    try
    {
        if (argc != 4)
        {
            throw std::runtime_error(std::string("Usage: ") + argv[0] + " <cpu|cuda> <input image> <x,y,w,h> ");
        }

        std::string strBackend       = argv[1];
        std::string strInputFileName = argv[2];

        if (sscanf(argv[3], "%d,%d,%d,%d", &originX, &originY, &templWidth, &templHeight) != 4)
        {
            throw std::runtime_error(
                "Invalid data format. Specify the bounding box of the input image as template image.");
        }

        // Load the input image
        cvImage = cv::imread(strInputFileName);
        if (cvImage.empty())
        {
            throw std::runtime_error("Can't open '" + strInputFileName + "'");
        }

        assert(cvImage.type() == CV_8UC3);

        // convert image to gray scale
        cvtColor(cvImage, cvImageU8, cv::COLOR_BGR2GRAY);

        if (originX + templWidth > cvImage.cols || originY + templHeight > cvImage.rows)
        {
            throw std::runtime_error("Bounding box is out of range of input image size");
        }

        cv::Rect templROI(originX, originY, templWidth, templHeight);
        cv::Mat croppedRef(cvImageU8, templROI);
        croppedRef.copyTo(cvTempl);

        // Now parse the backend
        VPIBackend backend;

        if (strBackend == "cpu")
        {
            backend = VPI_BACKEND_CPU;
        }
        else if (strBackend == "cuda")
        {
            backend = VPI_BACKEND_CUDA;
        }
        else
        {
            throw std::runtime_error("Backend '" + strBackend + "' not recognized, it must be either cpu, cuda");
        }

        // 1. Initialization phase ---------------------------------------

        CHECK_STATUS(vpiStreamCreate(backend, &stream));

        // We now wrap the loaded image into a VPIImage object to be used by VPI.
        // VPI won't make a copy of it, so the original
        // image must be in scope at all times.
        CHECK_STATUS(vpiImageCreateWrapperOpenCVMat(cvImageU8, 0, &input));

        // Create template iamge
        CHECK_STATUS(vpiImageCreateWrapperOpenCVMat(cvTempl, 0, &templ));

        // Now create the output image.
        outWidth  = cvImage.cols - templWidth + 1;
        outHeight = cvImage.rows - templHeight + 1;
        CHECK_STATUS(vpiImageCreate(outWidth, outHeight, VPI_IMAGE_FORMAT_F32, 0, &output));
        CHECK_STATUS(vpiImageCreate(outWidth, outHeight, VPI_IMAGE_FORMAT_F32, 0, &outputScaled));
        CHECK_STATUS(vpiImageCreate(outWidth, outHeight, VPI_IMAGE_FORMAT_U8, 0, &outputU8));

        CHECK_STATUS(vpiArrayCreate(10000, VPI_ARRAY_TYPE_KEYPOINT_F32, 0, &minCoords));
        CHECK_STATUS(vpiArrayCreate(10000, VPI_ARRAY_TYPE_KEYPOINT_F32, 0, &maxCoords));

        // Create payload
        CHECK_STATUS(vpiCreateTemplateMatching(backend, cvImage.cols, cvImage.rows, &payload));

        CHECK_STATUS(vpiCreateMinMaxLoc(backend, cvImage.cols, cvImage.rows, VPI_IMAGE_FORMAT_F32, &payloadMinMax));

        // 2. Computation phase ---------------------------------------

        // Set source image
        CHECK_STATUS(vpiTemplateMatchingSetSourceImage(stream, backend, payload, input));

        // Set source image
        CHECK_STATUS(vpiTemplateMatchingSetTemplateImage(stream, backend, payload, templ, NULL));

        // Submit
        CHECK_STATUS(vpiSubmitTemplateMatching(stream, backend, payload, output, VPI_TEMPLATE_MATCHING_NCC));

        CHECK_STATUS(vpiSubmitMinMaxLoc(stream, backend, payloadMinMax, output, minCoords, maxCoords));

        // Convert output from F32 to U8
        VPIConvertImageFormatParams params;
        CHECK_STATUS(vpiInitConvertImageFormatParams(&params));
        params.scale = 255;
        CHECK_STATUS(vpiSubmitConvertImageFormat(stream, backend, output, outputScaled, &params));

        CHECK_STATUS(vpiSubmitConvertImageFormat(stream, backend, outputScaled, outputU8, NULL));

        // Wait until the algorithm finishes processing
        CHECK_STATUS(vpiStreamSync(stream));

        // Now let's retrieve the output image contents and output it to disk
        {
            // Lock output image to retrieve its data on cpu memory
            VPIArrayData maxCoordsData;
            CHECK_STATUS(vpiArrayLockData(maxCoords, VPI_LOCK_READ, VPI_ARRAY_BUFFER_HOST_AOS, &maxCoordsData));

            VPIKeypointF32 *max_coords = (VPIKeypointF32 *)maxCoordsData.buffer.aos.data;
            int max_i                  = max_coords[0].y;
            int max_j                  = max_coords[0].x;

            VPIImageData outData;
            CHECK_STATUS(vpiImageLockData(outputU8, VPI_LOCK_READ, VPI_IMAGE_BUFFER_HOST_PITCH_LINEAR, &outData));

            // Returned data consists of host-accessible memory buffers in pitch-linear layout.
            assert(outData.bufferType == VPI_IMAGE_BUFFER_HOST_PITCH_LINEAR);

            cv::Mat cvOut;
            CHECK_STATUS(vpiImageDataExportOpenCVMat(outData, &cvOut));

            cv::Rect rect(max_i - 20, max_j - 20, 40, 40);
            rectangle(cvOut, rect, 255, 2);
            imwrite("template_matching_score_" + strBackend + ".png", cvOut);

            printf("Provided coord of bounding box for the template is [%d, %d] with w=%d and h=%d \n", originY,
                   originX, templWidth, templHeight);
            printf("Template matching location coord is [%d, %d] \n", max_j, max_i);

            // Done handling output image, don't forget to unlock it.
            CHECK_STATUS(vpiArrayUnlock(maxCoords));
            CHECK_STATUS(vpiImageUnlock(outputU8));
        }
    }
    catch (std::exception &e)
    {
        std::cerr << e.what() << std::endl;
        retval = 1;
    }

    // Clean up

    // Make sure stream is synchronized before destroying the objects
    // that might still be in use.
    vpiStreamSync(stream);

    vpiPayloadDestroy(payload);
    vpiPayloadDestroy(payloadMinMax);

    vpiImageDestroy(input);
    vpiImageDestroy(templ);
    vpiImageDestroy(output);
    vpiImageDestroy(outputU8);
    vpiArrayDestroy(minCoords);
    vpiArrayDestroy(maxCoords);
    vpiStreamDestroy(stream);

    return retval;
}
