# Copyright (c) 2022, NVIDIA CORPORATION. All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#  * Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
#  * Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#  * Neither the name of NVIDIA CORPORATION nor the names of its
#    contributors may be used to endorse or promote products derived
#    from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS ``AS IS'' AND ANY
# EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
# PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
# PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
# OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import cv2
import sys
import vpi
import numpy as np
from PIL import Image, ImageOps
from argparse import ArgumentParser

# Parse command line arguments
parser = ArgumentParser()
parser.add_argument('backend', choices=['cpu','cuda'],
                    help='Backend to be used for processing')

parser.add_argument('s', metavar='filename',
                    help='Image to be used as source')

parser.add_argument('t', metavar='X,Y,W,H',
                    help='Use specified bounding box of source as template image')

args = parser.parse_args()

if args.backend == 'cpu':
    backend = vpi.Backend.CPU
elif args.backend == 'cuda':
    backend = vpi.Backend.CUDA
else:
    sys.exit("Un-supported backend")

try:
    templBbox = np.array([int(x) for x in args.t.split(',')])
except ValueError:
    exit("Error parsing template bounding box")

# Load input into a vpi.Image
try:
    srcData = np.asarray(ImageOps.grayscale(Image.open(args.s)))
except IOError:
    sys.exit("Source file not found")
except:
    sys.exit("Error with source file")

src = vpi.asimage(srcData)
templData = srcData[templBbox[1]:templBbox[1]+templBbox[3], templBbox[0]:templBbox[0]+templBbox[2]]
templ = vpi.asimage(templData)

# Using the chosen backend,
with backend:
    output = vpi.templateMatching(src, templ)

    temp = output.convert(vpi.Format.F32, backend=vpi.Backend.CUDA, scale = 255)

    min_coords, max_coords = temp.minmaxloc(min_capacity=10000, max_capacity=10000)

    output = temp.convert(vpi.Format.U8, backend=vpi.Backend.CUDA)

# Lock output
with max_coords.rlock_cpu() as max_data:
    max_loc = tuple(max_data[0].astype(int)[::-1])

# Save result to disk
print('Provided coord of bounding box for the template is [{}, {}] with w={} and h={}'.format(templBbox[1], templBbox[0], templBbox[2], templBbox[3]))
print('Template matching location coord is [{}, {}]'.format(max_loc[0], max_loc[1]))
outputData = output.cpu()
cv2.rectangle(outputData, (max_loc[1]-20, max_loc[0]-20), (max_loc[1]+20, max_loc[0]+20), 255, 2)
Image.fromarray(outputData).save('template_matching_score_python'+str(sys.version_info[0])+'_'+args.backend+'.png')
